try:
    import llama
except ImportError:
    raise ImportError("Please install llama to use Ollama LLM.")
from typing import List
from llama import Tokenizer
import time
import requests, json
import os
from .utils import match_model
from .base import BaseModel

def initialize_client():
    ollama_modelpath = os.path.join(os.path.expanduser('~'), ".ollama", "ollama_model.info")
    if os.environ.get("OLLAMA_URL"):
        tokenmodel_path = os.environ.get("OLLAMA_TOKENIZER_PATH")
        ollama_url = os.environ.get("OLLAMA_URL")
        return tokenmodel_path, ollama_url
    elif os.path.exists(ollama_modelpath):
        with open(ollama_modelpath, 'r') as f:
            lines = f.read().splitlines()
            if len(lines) >= 2:
                tokenmodel_path = lines[0].strip()
                ollama_url = lines[1].strip()
                return tokenmodel_path, ollama_url
        raise FileNotFoundError(
            """
            Please create a file at ~/.ollama/ollama_model.info with your Ollama tokenizer model path and Ollama URL.
            The first line should be your tokenizer model path and the second line should be your Ollama URL.
            """
        )
    else:
        raise FileNotFoundError(
            """
            Please set the environment variables OLLAMA_URL and OLLAMA_TOKENIZER_PATH or create a file at ~/.ollama/ollama_model.info with your Ollama tokenizer model path and Ollama URL.
            """
        )

class Ollama_chat(BaseModel):

    def __init__(self, model_name, global_counter=None):

        tokenmodel_path, ollama_url = initialize_client()
        model_max, enc_fn = match_model(model_name)
        model_name = model_name.replace('ollama-','')
        self.model_name = model_name
        self.name = model_name
        self.sp_model = Tokenizer(tokenmodel_path)
        self.url = f'{ollama_url}/api/chat'
        self.model_max = model_max
        super().__init__(model_name, global_counter, "chat")
    
    def encode(self, txt, bos: bool = True, eos: bool = False) -> List[int]:
        return self.sp_model.encode(txt,bos=bos,eos=eos)

    def decode(self, txt, bos: bool = True, eos: bool = False) -> str:
        return self.sp_model.decode(txt,bos,eos)

    def query_chat(self, messages, shrink_idx, max_gen=512, temp=0.):
        model_max = self.model_max
        messages = self.shrink_msg(messages, shrink_idx, model_max-max_gen)
        while(True):
            try:
                ollama_body = {"messages": messages, "model": self.name, "stream": False}
                completion = requests.post(self.url,json=ollama_body)
                print(ollama_body)
                response = json.loads(completion.content.decode('utf-8'))
                content = response['message']['content']

                completion_count = self.count_tokens(content)
                prompt_count = self.count_tokens(messages)
                total_tokens = prompt_count + completion_count
                return content, {"prompt":prompt_count, "completion":completion_count, "total":total_tokens}
            except Exception as e:
                e = str(e)
                if "However, your messages resulted in" in e:
                    print("error:", e)
                    index = e.find("your messages resulted in ")
                    import re
                    val = int(re.findall(r'\d+', e[index + len("your messages resulted in ") : ])[0])
                    index2 = e.find("maximum context length is ")
                    model_max = int(re.findall(r'\d+', e[index2 + len("maximum context length is "):])[0])
                    messages = self.shrink_msg_by(messages, shrink_idx, val-model_max)
                else:
                    time.sleep(5)
                    print(e)



token_counter = {
    "token_completion":{},
    "token_prompt":{},
    "api_calls":{}
}

def get_token_counts():
    global token_counter
    return token_counter

def query_gpt_chat(model):
    from .GPT import GPT_chat
    global token_counter
    query_model = GPT_chat(model, token_counter)
    return query_model

def query_claude_chat(model):
    from .claude import Claude_chat
    global token_counter
    query_model = Claude_chat(model, token_counter)
    return query_model

def query_ollama_chat(model):
    from .ollama import Ollama_chat
    global token_counter
    #model = model.replace('ollama-','')
    query_model = Ollama_chat(model, token_counter)
    return query_model

def get_query(LLM_name):
    """Get the query model for the specified LLM_name.

    Currently supported LLMs:

    - GPT-4

    - GPT-3.5

    - Claude

    - Ollama
    
    Args:
        LLM_name (str): Name of the LLM model
        tokenmodel_path (str): Path to the token model
    
    Returns:
        LLM_API_FUNCTION: The query model for the specified LLM_name
    """
    print("Warning: The built-in LLM API is only for testing, and will not be actively maintained. Please write your own API for production use.")
    if LLM_name.lower().startswith("gpt-4") or LLM_name.lower().startswith("gpt-3.5"):
        return query_gpt_chat(model=LLM_name)
    elif LLM_name.lower().startswith("claude"):
        return query_claude_chat(model=LLM_name)
    elif LLM_name.lower().startswith("ollama-"):
        return query_ollama_chat(model=LLM_name)
    else:
        raise NotImplementedError("LLM {} not implemented".format(LLM_name))
